//----------------------------------------------------------------------------
// 
//	file: spikelayer.cc
// 
//	function: extends netlayer -- make neurons spike 
//	author:   Yoonsuck Choe <yschoe@cs.utexas.edu>
//	creation: Mon Aug 24 02:13:55 CDT 1998
// 
//	$Revision: 1.9 $ $Date: 2001/11/30 20:24:15 $
//	$Log: spikelayer.cc,v $
//	Revision 1.9  2001/11/30 20:24:15  yschoe
//	pglissom v1.0 release
//
//	Revision 1.8  2000/01/30 14:56:21  yschoe
//	*** empty log message ***
//
//	Revision 1.7  1999/12/22 05:29:48  yschoe
//	activation function applied to membrane potential
//
//	Revision 1.6  1998/11/17 12:44:56  yschoe
//	1. fixed activation to exactly match gaus-slissom 97fall
//	2. added print_stat() for neuron stat gathering
//
//	Revision 1.5  1998/11/01 08:54:33  yschoe
//	*** empty log message ***
//
//	Revision 1.4  1998/10/26 03:58:20  yschoe
//	1. relocated reset_matrix() to appropriate positions where it is
//	   immediately necessary
//	2. calc_spike() moved to the beginning of the activate() function,
//	   as it was done in gaus-slissom/98jun25/
//	3. output = avg_spiking in learn(). This was necessary so that it is
//	   the same as gaus-slissom/98jun25
//
//	Revision 1.3  1998/10/16 10:08:04  yschoe
//	attempting to fix bug and replicate previous results.
//	activate() and calls within have been updated.
//
//	Revision 1.2  1998/10/05 19:48:46  yschoe
//	1. added get_inh|exc|affw_matrix() functions
//	2. added setup_weight() functions
//	3. separated calc_affresp and calc_latresp from calc_netinput
//	4. set_matrix() for gaussian weight setup
//
//	Revision 1.1  1998/09/04 03:36:37  yschoe
//	fixed all warnings generated by -Wall
//
//
//----------------------------------------------------------------------------

#include "spikelayer.h"


//----------------------------------------------------------------------------
//	spikelayer() 
//	  - Constructor
//----------------------------------------------------------------------------
spikelayer::spikelayer(char *name, int cols, int rows, int exc_rad, int inh_rad,
                 int max_inp_layers, netlayer *feeding,
                 int inp_cols, int inp_rows, int  aff_rad,
		 REAL syn_decay, REAL inh_decay, 
		 REAL theta_arg, REAL tau_arg, int padded_flag)
	: netlayer(name,cols,rows,exc_rad,inh_rad,max_inp_layers,feeding,
		inp_cols,inp_rows,aff_rad,padded_flag) {

  // 1. assign newly introduced parameters
  spikelayer::synapse_decay = syn_decay;
  spikelayer::inhloop_decay = inh_decay;
  spikelayer::theta	    = theta_arg;
  spikelayer::tau	    = tau_arg;

  // 2. allocate newly introduced matrices
  spike	       = alloc_matrix(cols, rows, spike);
  spikeact_q   = alloc_matrix(cols, rows, spikeact_q);
  inhloop_q    = alloc_matrix(cols, rows, inhloop_q);
  avg_spiking  = alloc_matrix(cols, rows); // REAL type matrix
  lfp 	       = alloc_matrix(cols, rows); // REAL type matrix

  	// 00jun6 additions - for spikelayer2_2
  new_s_exc 	= alloc_matrix(cols, rows);
  new_s_inh	= alloc_matrix(cols, rows);
  new_s_int	= alloc_matrix(cols, rows);
  s_exc		= alloc_matrix(cols, rows);
  s_inh		= alloc_matrix(cols, rows);
  s_int		= alloc_matrix(cols, rows);
  new_eta	= alloc_matrix(cols, rows);
  eta		= alloc_matrix(cols, rows);

  new_thresh	= alloc_matrix(cols, rows);
  thresh	= alloc_matrix(cols, rows);
  v		= alloc_matrix(cols, rows);
  r		= alloc_matrix(cols, rows);
  new_r		= alloc_matrix(cols, rows);

  // 3. allocate and initlialize look-up-table arrays 
  synapse_lut  = init_leaky_table(SPIKE_QLEN,synapse_decay);
  inhloop_lut  = init_leaky_table(SPIKE_QLEN,inhloop_decay);

  // 4.1 add new matrices to the blackboard
  // todo : must add :  net_bb->add("spike",spike,cols,rows);
  net_bb->add("spikeact_q",spikeact_q,cols,rows);
  net_bb->add("inhloop_q",inhloop_q,cols,rows);
  // added 00oct2
  net_bb->add("avg_spiking",avg_spiking,cols,rows);
  net_bb->add("spike",spike,cols,rows);

  // 4.2 add new variables to the blackboard
  net_bb->add("synapse_decay",&synapse_decay);
  net_bb->add("inhloop_decay",&inhloop_decay);
  net_bb->add("theta",&theta);
  net_bb->add("tau",&tau);
  net_bb->add("auto_theta_flag",&auto_theta_flag);
  net_bb->add("auto_theta_rate",&auto_theta_rate);

  net_bb->add("hard_refractory",&hard_refractory);
  net_bb->add("init_eta_randomness",&init_eta_randomness);

  // 5. set up bit-count LUT (for faster avg spiking rate calculation)
  REAL bitcount_tablelen = (REAL)pow(2.0,SPIKE_HISTQLEN);
  bitcount_lut = new REAL [(int)bitcount_tablelen];
  for (int i=0; i<bitcount_tablelen; ++i) {
	bitcount_lut[i] = 0;
	int q = i;
        for (int bit=0; bit<SPIKE_HISTQLEN; ++bit) {
		bitcount_lut[i] += (REAL)(q & 0x1);
		q >>= 1;
	}
	// now get an average spiking rate
	bitcount_lut[i] /= (REAL) SPIKE_HISTQLEN;
	// debug : printf("%x : %f\n",i,bitcount_lut[i]);
  } 
  

  // 6. Clear matrices
  reset_matrix(output,cols,rows);
  reset_matrix(net_input,cols,rows);
  reset_matrix(net_aff_input,cols,rows);
  reset_matrix(net_exc_input,cols,rows);
  reset_matrix(net_inh_input,cols,rows);
  reset_matrix(spike,cols,rows);
  reset_matrix(spikeact_q,cols,rows);
  reset_matrix(avg_spiking,cols,rows);
  reset_matrix(lfp,cols,rows);

  // 7. defaults : for backward compatibility
  init_eta_randomness = 0.0;
}

//----------------------------------------------------------------------------
//	~spikelayer() 
//	  - Destructor
//----------------------------------------------------------------------------
spikelayer::~spikelayer() {

  // 1. delete allocated matrices
  delete_matrix(cols, rows, spike);
  delete_matrix(cols, rows, spikeact_q);
  delete_matrix(cols, rows, inhloop_q);

  // 2. delete allcated arrays
  delete [] synapse_lut;
  delete [] inhloop_lut;

}

//----------------------------------------------------------------------------
//	activate() 
//	  - do mostly as in netlayer.cc except for spiking stuff
//----------------------------------------------------------------------------
void spikelayer::activate(int calc_aff) {
  
  // excerpt from gaus-slissom:leaky.c
  //
  // - Clear queues (when training only)
  // - Generate input
  // - Settling : aff contrib MUST be calculated everytime
  //	- calculate activity
  //		* aff_sum += wgt[i][j] * input->output[i][j]
  //		* lat_sum += wgt[i][j] * output->output[i][j]]
  //	- generate spike if net_input[i][j]>theta+tau*inhloop_lut[i][j]
  //		(note: sigmoid was not used here??)
  //		* advance spikes in queues
  //		* add new spike to queues
  // - Modify weights
  //	- history (no decay) : update weights according to
  //		sigmoid(average spiking rate) <-- this is strange. Fix it.

  // Goal:
  // (1) calculate spike, (2) push spike, then finally (3) 
  // put LUT value in this->output

  // 0. calc and push spikes
  push_spike();

  // 1. use netlayer::calc_affresp() to calculate net_input.
  // 	todo: If we assume that input spikes reach a stable firing before
  //	feeding into the cortex, we can avoid calculating affresp 
  //	everytime during settling.
  calc_affresp(); 	// net_aff_input <-- wgt * (feeding->output); 

  // 2. use netlayer::calc_latresp() to calculate lateral contrib
  calc_latresp(); 	// net_exc/inh_input <- wgt*output (from previous step)
  calc_contrib_sum(0);  // net_input <-- net_aff_input + net_exc/inh_input

  // 3. calculate spike
  calc_spike();		// output <-- shifted, decayed sum of spikes

}


//----------------------------------------------------------------------------
//	push_spike() 
//	  - push current spike into spike_q
//----------------------------------------------------------------------------
void spikelayer::push_spike() {

  for (int i=0; i<cols; ++i) {
     for (int j=0; j<rows; ++j) {

	// 1. push spike : do not mask it since it will be 
	//    used to calculate avg spiking rate too.
	spikeact_q[i][j] = (spikeact_q[i][j] << 1) | spike[i][j];

  	// 2. put LUT value in this->output (decayed sum of spikes)
	// 	this is used in the next step of learning
  	// *. at this point output[i][j] has the same semantics as the q_sum
	//    in the previous implementation of slissom.
	// *. spikeact_q is identical to inhloop_q, so there's
 	//    no need to push spikes into the inhloop_q.
	output[i][j] = act_func(synapse_lut[(unsigned)spikeact_q[i][j] & SPIKE_QMASK ]);

 	avg_spiking[i][j] = act_func(
		bitcount_lut[(unsigned)spikeact_q[i][j] & SPIKE_HISTQMASK] 
		);
     }
  }

}

//----------------------------------------------------------------------------
//	calc_spike() 
//	  - generate spike and push spike, put LUT value, etc.
//----------------------------------------------------------------------------
void spikelayer::calc_spike() {

  for (int i=0; i<cols; ++i) {
     for (int j=0; j<rows; ++j) {

	// 0. just for displaying -- never used in calculation 
	lfp[i][j] = act_func(net_input[i][j]);

  	// 1. generate spike
	// *. spikeact_q is identical to inhloop_q, so there's
 	//    spikeact_q is used instead of inhloop_q when looking up the
 	//    inhloop_lut.
	spike[i][j] = ( net_input[i][j] 
		> theta 
		  +(tau*inhloop_lut[(unsigned)spikeact_q[i][j] & SPIKE_QMASK])
	    ) ? 1 : 0;
	/* DEBUG
	if (i==2 && j==2) {
		disp_binary(spikeact_q[i][j] & SPIKE_HISTQMASK);
		cout << "  :  " 
		     << ((unsigned)spikeact_q[i][j] & SPIKE_HISTQMASK) ;
	}*/

     }
  }

}


//----------------------------------------------------------------------------
//	learn() 
//	  - calculate avg spiking rate and then call netlayer::learn()
//----------------------------------------------------------------------------
void spikelayer::learn() {

  
  // debug: cout << "spikelayer::learn()\n";

  // 1. Now output[i][j] takes on a different context -- avg spiking rate
  //    -- for weight adaptation
  for (int i=0; i<cols; ++i) {
     for (int j=0; j<rows; ++j) {
	// Option 1. avg spiking rate 
	
   	output[i][j] = act_func(
		bitcount_lut[(unsigned)spikeact_q[i][j] & SPIKE_HISTQMASK] 
		);
     //output[i][j] = avg_spiking[i][j];

	// Option 2. membrane potential
	// leave output[i][j] as is.
   	// output[i][j] = lfp[i][j];
     }
  }

  // 2. then actually adapt weights
  //	netlayer::learn() uses output[i][j] for learing
  netlayer::learn();

}
//----------------------------------------------------------------------------
//	print_stat() 
//	  - print status of (col,row) neuron.
//----------------------------------------------------------------------------
void spikelayer::print_stat(int col, int row) {

  cout 	<< net_aff_input[col][row] << " "
	<< (lat_exc_contri*net_exc_input[col][row]) << " "
	<< (lat_inh_contri*net_inh_input[col][row]) << " "
	<< net_input[col][row] << " "
	<< (synapse_lut[(unsigned)spikeact_q[col][row]&SPIKE_QMASK ]) << " "
	<< theta << " "
	<< (theta+tau*inhloop_lut[(unsigned)spikeact_q[col][row]&SPIKE_QMASK])<<endl;

}

//----------------------------------------------------------------------------
//	print_stat() 
//	  - print status of (col,row) neuron.
//----------------------------------------------------------------------------
void spikelayer::print_stat(int cur_inp, int inp_col, int inp_row) {

  int col = (int)(inp_col * cols / (REAL)feedings[cur_inp]->cols);
  int row = (int)(inp_row * cols / (REAL)feedings[cur_inp]->rows);

  if (col >= cols || row >= rows) {
	cerr << "spikelayer::print_stat - index out of range\n";
	return;
  }

  cout 	<< net_aff_input[col][row] << " "
	<< (lat_exc_contri*net_exc_input[col][row]) << " "
	<< (lat_inh_contri*net_inh_input[col][row]) << " "
	<< net_input[col][row] << " "
	<< (synapse_lut[(unsigned)spikeact_q[col][row]&SPIKE_QMASK ]) << " "
	<< theta << " "
	<< (theta+tau*inhloop_lut[(unsigned)spikeact_q[col][row]&SPIKE_QMASK])<<endl;

}

